import matplotlib.pyplot as plt
import muon as mu
import numpy as np
import os
import pyro
import random
import torch
from scipy.optimize import linear_sum_assignment
from sklearn.metrics import confusion_matrix
from sklearn.neighbors import NearestNeighbors

def set_seed(seed=9):
    """Initializes uniform random number generators.

    Parameters
    ----------
        seed : int, default=9
            Optional.
            Value to be used for all random number generation.

    Returns
    -------
        None
    """

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True
    pyro.set_rng_seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)

def map_labels(X_labels, Y_labels):
    """Maps predicted clustering labels to the given ground truth using linear
    sum assignment.
    
    Parameters
    ----------
        X_labels : numpy.ndarray, shape=(n_samples,)
            Ground truth cluster labels.
        Y_labels : numpy.ndarray, shape=(n_samples,)
            Predicted cluster labels.

    Returns
    -------
        labels : numpy.ndarray, shape=(n_samples,)
            TODO
    """

    scores = confusion_matrix(Y_labels, X_labels)
    row, col = linear_sum_assignment(scores, maximize=True)
    labels = np.zeros_like(X_labels)

    for i in row:
        labels[Y_labels == i] = col[i]

    return labels

def get_spatial(mdata, spatial_key='spatial', modality_key='morphological'):
    """TODO
    
    Parameters
    ----------
        mdata : TODO
            TODO
        spatial_key : str, default='spatial'
            Optional.
            TODO
        modality_key : str, default='physical'
            Optional.
            TODO
    
    Returns
    -------
        x : numpy.ndarray, shape=(n_samples,)
            TODO
        y : numpy.ndarray, shape=(n_samples,)
            TODO
    """

    try:
        x, y = mdata[modality_key].obsm[spatial_key].T
    except KeyError:
        x, y = mdata.obsm[spatial_key].T

    return x, y
    
def get_features(mdata, feature_key='protein', imagenet_key='imagenet'):
    """TODO

    Parameters
    ----------
        mdata : TODO
            TODO
        feature_key : str, default='protein'
            Optional.
            TODO
        imagenet_key : str, default='imagenet'
            Optional.
            TODO
    
    Returns
    -------
        features : numpy.ndarray, shape=(n_samples, n_features)
            TODO
    """

    try:
        features = mdata[feature_key].X
    except KeyError:
        features = mdata.X

    if imagenet_key:
        imagenet = mdata[imagenet_key].X
        features = np.hstack([features, imagenet])

    return features
    
def get_labels(mdata, label_key='celltype', modality_key='protein'):
    """TODO
    
    Parameters
    ----------
        mdata : TODO
            TODO
        label_key : str, default='celltype'
            Optional.
            TODO
        modality_key : str, default='protein'
            Optional.
            TODO

    Returns
    -------
        labels : numpy.ndarray, shape=(n_samples,)
            TODO
    """
    
    try:
        labels = mdata[modality_key].obs[label_key]
    except KeyError:
        labels = mdata.obs[label_key]

    return labels

def remove_lonely(data, labels, n_neighbors=12, threshold=225.):
    """Filters out samples that are too far removed from the data of interest
    according to the given threshold.

    Parameters
    ----------
        data : numpy.ndarray, shape=(n_samples, n_features)
            Collection of samples to be filtered.
        labels : numpy.ndarray, shape=(n_samples,)
            Component labels for each sample.
        n_neighbors : int, default=12
            Optional.
            TODO
        threshold : float, default=225.0
            Optional.
            TODO

    Returns
    -------
        data : numpy.ndarray, shape=(n_samples, n_features)
            TODO
        labels : numpy.ndarray, shape=(n_samples,)
            TODO
    """

    knn = NearestNeighbors(n_neighbors=n_neighbors).fit(data[:, :2])
    max_dist = knn.kneighbors()[0].max(-1)
    mask_idx, = np.where(max_dist > threshold)
    data = np.delete(data, mask_idx, axis=0)
    labels = np.delete(labels, mask_idx, axis=0)

    return data, labels

def read_anndata(filename, spatial_key='spatial', feature_key='protein', spatial_modality='morphological', imagenet_key='imagenet', label_key='celltype', label_modality='protein', n_neighbors=12, threshold=225., delineate=True, tensor=False):
    """Reads the specified annotated data object according to the format used by
    the NYGC Technology Innovation Laboratory.
    
    Parameters
    ----------
        filename : str
            TODO
        spatial_key : str, default='spatial'
            Optional.
            TODO
        spatial_modality : str, default='morphological'
            Optional.
            TODO
        feature_key : str, default='protein'
            Optional.
            TODO
        imagenet_key : str, default='imagenet'
            Optional.
            TODO
        label_key : str, default='protein:celltype'
            Optional.
            TODO
        label_modality : str, default='protein'
            Optional.
            TODO
        n_neighbors : int, default=12
            Optional.
            TODO
        threshold : float, default=225.0
            Optional.
            TODO
        delineate : bool, default=True
            Optional.
            TODO
        tensor : bool, default=False
            Optional.
            TODO

    Returns
    -------
        data : numpy.ndarray|torch.Tensor, shape=(n_samples, n_features)
            TODO
        labels : numpy.ndarray|torch.Tensor, shape=(n_samples,)
            TODO
    """

    mdata = mu.read(filename)
    x, y = get_spatial(mdata, spatial_key, spatial_modality)
    features = get_features(mdata, feature_key, imagenet_key)
    data = np.hstack([x[None].T, y[None].T, features])
    labels = get_labels(mdata, label_key, label_modality)
    _, labels = np.unique(labels, return_inverse=True)

    if threshold is not None:
        data, labels = remove_lonely(data, labels, n_neighbors, threshold)

    if delineate:
        data = np.hstack([np.zeros((data.shape[0], 1)), data])

    if tensor:
        return torch.tensor(data, dtype=torch.float32), torch.tensor(labels, dtype=torch.int32)
    return data, labels

def read_anndatas(filenames, spatial_key='spatial', spatial_modality='morphological', feature_key='protein', imagenet_key='imagenet', label_key='celltype', label_modality='protein', n_neighbors=12, threshold=225., delineate=False, tensor=False):
    """Reads the specified annotated data object according to the format used by
    the NYGC Technology Innovation Laboratory.
    
    Parameters
    ----------
        filenames : list
            TODO
        spatial_key : str, default='spatial'
            Optional.
            TODO
        spatial_modality : str, default='morphological'
            Optional.
            TODO
        feature_key : str, default='protein'
            Optional.
            TODO
        imagenet_key : str, default='imagenet'
            Optional.
            TODO
        label_key : str, default='celltype'
            Optional.
            TODO
        label_modality : str, default='protein'
            Optional.
            TODO
        n_neighbors : int, default=12
            Optional.
            TODO
        threshold : float, default=225.0
            Optional.
            TODO
        delineate : bool, default=False
            Optional.
            TODO
        tensor : bool, default=False
            Optional.
            TODO

    Returns
    -------
        data : numpy.ndarray|torch.Tensor, shape=(n_samples, n_features)
            TODO
        labels : numpy.ndarray|torch.Tensor, shape=(n_samples,)
            TODO
    """

    data, labels = [], []

    for i, filename in enumerate(filenames):
        mdata = mu.read(filename)
        x, y = get_spatial(mdata, spatial_key, spatial_modality)
        features = get_features(mdata, feature_key, imagenet_key)
        data_i = np.hstack([x[None].T, y[None].T, features])
        labels_i = get_labels(mdata, label_key, label_modality)
        _, labels_i = np.unique(labels_i, return_inverse=True)

        if threshold is not None:
            data_i, labels_i = remove_lonely(data_i, labels_i, n_neighbors, threshold)

        data.append(np.hstack([i*np.ones((data_i.shape[0], 1)), data_i]))
        labels.append(labels_i)

    data, labels = np.vstack(data), np.hstack(labels)

    if tensor:
        return torch.tensor(data, dtype=torch.float32), torch.tensor(labels, dtype=torch.int32)
    return data, labels

def itemize(n, *items):
    """TODO
    
    Parameters
    ----------
        n : int
            TODO
        items : object
            TODO
    
    Yields
    ------
        items : tuple, shape=(n,)
            TODO
    """

    for i in items:
        yield i if isinstance(i, (list, tuple)) else (i,)*n

def format(ax, aspect='equal', show_ax=True):
    """TODO
    
    Parameters
    ----------
        ax : matplotlib.axis
            TODO
        aspect : str, default='equal'
            Optional.
            TODO
        show_ax : bool, default=True
            Optional.
            TODO

    Returns
    -------
        ax : matplotlib.axis
            TODO
    """

    ax.set_aspect(aspect)
    
    if not show_ax:
        ax.axis('off')

    return ax

def show_dataset(locs, labels, size=15, figsize=5, show_ax=True, colormap='tab10', filename=None):
    """TODO
    
    Parameters
    ----------
        locs : numpy.ndarray, shape=(n_samples, 2)
            TODO
        labels : numpy.ndarray, shape=(n_samples,)
            TODO
        size : int, default=15
            Optional.
            TODO
        show_ax : bool, default=True
            Optional.
            TODO
        figsize : int, default=10
            Optional.
            TODO
        colormap : str, default='tab20'
            Optional.
            TODO
        filename : str, default=None
            Optional.
            TODO

    Returns
    -------
        None
    """

    figsize, = itemize(2, figsize)
    fig, ax = plt.subplots(figsize=figsize)
    locs = locs[:, -2:].T
    color = plt.colormaps[colormap](labels)
    ax.scatter(*locs, s=size, c=color)
    ax = format(ax, aspect='equal', show_ax=show_ax)

    if filename is not None:
        fig.savefig(filename, bbox_inches='tight')

def show_datasets(locs, labels, size=15, figsize=10, show_ax=True, colormap='tab10', filename=None):
    """TODO
    
    Parameters
    ----------
        locs : numpy.ndarray, shape=(n_samples, 3)
            TODO
        labels : numpy.ndarray, shape=(n_samples,)
            TODO
        size : int, default=15
            Optional.
            TODO
        figsize : int, default=10
            Optional.
            TODO
        show_ax : bool, default=True
            Optional.
            TODO
        colormap : str, default='tab20'
            Optional.
            TODO
        filename : str, default=None
            Optional.
            TODO

    Returns
    -------
        TODO
    """
    
    n_datasets = np.unique(locs[:, 0]).shape[0]
    size, = itemize(n_datasets, size)
    figsize, = itemize(2, figsize)
    fig, ax = plt.subplots(1, n_datasets, figsize=figsize)
    axes = (ax,) if n_datasets == 1 else ax

    for i in range(n_datasets):
        idx = locs[:, 0].astype(np.int32) == i
        idx_locs = locs[idx, -2:].T
        color = plt.colormaps[colormap](labels[idx])
        axes[i].scatter(*idx_locs, s=size[i], c=color)
        format(axes[i], aspect='equal', show_ax=show_ax)

    if filename is not None:
        fig.savefig(filename, bbox_inches='tight')

    return fig, axes

def show_logs(*logs, size=15, figsize=10, show_ax=True, filename=None):
    """TODO
    
    Parameters
    ----------
        logs : tuple, shape=(n_logs,)
            TODO
        size : int, default=15
            Optional.
            TODO
        figsize : int, default=10
            Optional.
            TODO
        show_ax : bool, default=True
            Optional.
            TODO
        filename : str, default=None
            Optional.
            TODO

    Returns
    -------
        TODO
    """

    n_logs = len(logs)
    size, = itemize(n_logs, size)
    figsize, = itemize(2, figsize)
    fig, ax = plt.subplots(1, n_logs, figsize=figsize)
    axes = (ax,) if n_logs == 1 else ax

    for idx, log in enumerate(logs):
        x = np.arange(len(log))
        axes[idx].plot(x, log)
        format(axes[idx], aspect='equal', show_ax=show_ax)

    if filename is not None:
        fig.savefig(filename, bbox_inches='tight')

    return fig, axes

def show_logs(*logs, figsize=5, show_ax=True, filename=None):
    """TODO
    
    Parameters
    ----------
        logs : tuple, shape=(n_logs,)
            TODO
        figsize : int, default=10
            Optional.
            TODO
        show_ax : bool, default=True
            Optional.
            TODO
        filename : str, default=None
            Optional.
            TODO

    Returns
    -------
        TODO
    """

    n_logs = len(logs)
    figsize, = itemize(2, figsize)
    fig, ax = plt.subplots(1, n_logs, figsize=figsize)
    axes = (ax,) if n_logs == 1 else ax

    for idx, log in enumerate(logs):
        if len(log) == 2:
            axes[idx].set_title(log[0])
            log = log[1]

        x = np.arange(len(log))
        axes[idx].plot(x, log)
        axes[idx].set_box_aspect(1)

        if not show_ax:
            axes[idx].tick_params(left=False, labelleft=False, bottom=False, labelbottom=False)

    if filename is not None:
        fig.savefig(filename, bbox_inches='tight')

    return fig, axes



import numpy as np
import torch
from scipy.cluster.vq import kmeans, vq
from scipy.spatial.distance import cdist
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.metrics import silhouette_score
from tqdm import tqdm
from .nets import VAE
from .util import set_seed

class SLDA(BaseEstimator, TransformerMixin):
    """TODO
    
    Parameters
    ----------
        n_topics : int
            TODO
        n_docs : int, default=150
            Optional.
            TODO
        vocab_size : int, default=25
            Optional.
            TODO
        vocab_steps : int, default=10
            Optional.
            TODO
        sigma : float, default=0.1
            Optional.
            TODO
        alpha : float, default=None
            Optional.
            TODO
        beta : float, default=None
            Optional.
            TODO
        batch_size : int, default=None
            Optional.
            TODO
        seed : int, default=9
            Optional.
            TODO

    Attributes
    ----------
        corpus_ : numpy.ndarray, shape=(n_samples, 6)
            TODO
        doc_locs_ : numpy.ndarray, shape=(n_docs, 3)
            TODO
        doc_topic_counts_ : numpy.ndarray, shape=(n_docs, n_topics)
            TODO
        topic_word_counts_ : numpy.ndarray, shape=(n_topics, vocab_size)
            TODO
        likelihood_log_ : list, shape=(n_steps,)
            TODO
    """

    def __init__(self, n_topics, n_docs=150, vocab_size=25, vocab_steps=10, sigma=.1, alpha=None, beta=None, batch_size=None, seed=9):
        super().__init__()

        if seed is not None:
            set_seed(seed)

        self.n_topics = n_topics
        self.n_docs = n_docs
        self.vocab_size = vocab_size
        self.vocab_steps = vocab_steps
        self.sigma = sigma
        self.alpha = 1/n_docs if alpha is None else alpha
        self.beta = 1/n_topics if beta is None else beta
        self.batch_size = batch_size

        self.corpus_ = None
        self.doc_locs_ = None
        self.doc_topic_counts_ = None
        self.topic_word_counts_ = None
        self.likelihood_log_ = []
        self.silhouette_log_ = []
        self.patience_log_ = []

    def _disp(self, idx, desc='SLDA'):
        """TODO
        
        Parameters
        ----------
            idx : int
                TODO
            desc : str, default='SLDA'
                TODO

        Returns
        -------
            disp : str
                TODO
        """

        scores = self.likelihood_log_[idx], self.patience_log_[idx]
        disp = f'{desc} | step {idx} | likelihood: {scores[0]} | patience: {scores[1]}'

        return disp

    def _featurize(self, locs, markers):
        """TODO
        
        Parameters
        ----------
            locs : numpy.ndarray, shape=(n_samples, 3)
                TODO
            markers : numpy.ndarray, shape=(n_samples, n_features)
                TODO

        Returns
        -------
            features : numpy.ndarray, shape=(n_samples, n_features)
                TODO
        """
        
        n_imgs = np.unique(locs[:, 0]).shape[0]
        features = []

        for i in range(n_imgs):
            idx = locs[:, 0] == i
            dists = cdist(locs[idx, 1:], locs[idx, 1:])
            weights = np.exp(-(dists/self.sigma)**2)
            features.append(weights.T@markers[idx])

        features = np.vstack(features)

        return features

    def _shuffle(self, words):
        """TODO
        
        Parameters
        ----------
            words : numpy.ndarray, shape=(n_samples,)
                TODO

        Returns
        -------
            docs : numpy.ndarray, shape=(n_samples,)
                TODO
            topics : numpy.ndarray, shape=(n_samples,)
                TODO
        """
        
        docs = np.random.choice(self.n_docs, (words.shape[0], 1))
        topics = np.random.choice(self.n_topics, (words.shape[0], 1))
        doc_range, topic_range = np.arange(self.n_docs), np.arange(self.n_topics)
        self.doc_topic_counts_ = (docs == doc_range).T@np.eye(self.n_topics)[topics.T[0]]
        self.topic_word_counts_ = (topics == topic_range).T@np.eye(self.vocab_size)[words]

        return docs, topics
    
    def _build(self, data):
        """TODO
        
        Parameters
        ----------
            data : numpy.ndarray, shape=(n_samples, n_features)
                TODO

        Returns
        -------
            corpus_ : numpy.ndarray, shape=(n_samples, 6)
                TODO
        """

        locs, markers = data[:, :3], data[:, 3:]
        self.n_docs = np.unique(locs[:, 0]).shape[0]*self.n_docs
        doc_idx = np.random.permutation(data.shape[0])[:self.n_docs]
        self.doc_locs_ = locs[doc_idx]
        features = self._featurize(locs, markers)
        codebook, _ = kmeans(features, self.vocab_size, self.vocab_steps)
        words, _ = vq(features, codebook)
        docs, topics = self._shuffle(words)
        self.corpus_ = np.hstack([locs, words[None].T, docs, topics])

        if self.batch_size is None:
            self.batch_size = self.corpus_.shape[0]

        return self.corpus_
    
    def _patience(self, idx, window):
        """TODO
        
        Parameters
        ----------
            window : int
                TODO

        Returns
        -------
            patience : bool
                TODO
        """

        if idx%window == 0 and idx > 2*window:
            return sum(self.likelihood_log_[-window:]) <= sum(self.likelihood_log_[-2*window:-window])
        return False
    
    def _sample_doc(self, loc, topic, return_likelihood=False):
        """TODO
        
        Parameters
        ----------
            loc : numpy.ndarray, shape=(3,)
                TODO
            topic : int
                TODO
            return_likelihood : bool, default=False
                Optional.
                TODO

        Returns
        -------
            doc : int
                TODO
            likelihood : float
                TODO
        """

        mask = (self.doc_locs_[:, 0] == loc[0]).astype(np.int32)
        doc_probs = mask*self.sigma**2/(((loc[1:] - self.doc_locs_[:, 1:])**2).sum(-1) + 1e-100)
        topic_probs = self.doc_topic_counts_[:, topic] + self.alpha
        topic_probs /= (self.doc_topic_counts_ + self.alpha).sum(-1)
        probs = doc_probs*topic_probs/(doc_probs*topic_probs).sum()
        doc = np.random.choice(self.n_docs, p=probs)
        
        if return_likelihood:
            return doc, probs[doc]
        
        return doc
    
    def _sample_topic(self, word, doc, return_likelihood=False, maximize=False):
        """TODO
        
        Parameters
        ----------
            word : int
                TODO
            doc : int
                TODO
            return_likelihood : bool, default=False
                Optional.
                TODO
            maximize : bool, default=False
                Optional.
                TODO

        Returns
        -------
            topic : int
                TODO
            likelihood : float
                TODO
        """

        topic_probs = self.doc_topic_counts_[doc] + self.alpha
        topic_probs /= (self.doc_topic_counts_[doc] + self.alpha).sum()
        word_probs = self.topic_word_counts_[:, word] + self.beta
        word_probs /= (self.topic_word_counts_ + self.beta).sum(-1)
        probs = topic_probs*word_probs/(topic_probs*word_probs).sum()
        topic = np.argmax(probs) if maximize else np.random.choice(self.n_topics, p=probs)

        if return_likelihood:
            return topic, probs[topic]
        
        return topic
    
    def _sample(self, loc, word, old_doc, old_topic, return_likelihood=False, maximize=False):
        """TODO
        
        Parameters
        ----------
            loc : numpy.ndarray, shape=(3,)
                TODO
            word : int
                TODO
            old_doc : int
                TODO
            old_topic : int
                TODO
            return_likelihood : bool, default=False
                Optional.
                TODO
            maximize : bool, default=False
                Optional.
                TODO

        Returns
        -------
            new_doc : int
                TODO
            new_topic : int
                TODO
            likelihood : float
                TODO
        """
        
        new_doc, doc_likelihood = self._sample_doc(loc, old_topic, return_likelihood)
        new_topic, topic_likelihood = self._sample_topic(word, old_doc, return_likelihood, maximize)
        
        if return_likelihood:
            return new_doc, new_topic, doc_likelihood + topic_likelihood

        return new_doc, new_topic
    
    def _decrement(self, word, doc, topic):
        """TODO
        
        Parameters
        ----------
            word : int
                TODO
            doc : int
                TODO
            topic : int
                TODO

        Returns
        -------
            doc_topic_counts_ : numpy.ndarray, shape=(n_docs, n_topics)
                TODO
            topic_word_counts_ : numpy.ndarray, shape=(n_topics, vocab_size)
                TODO
        """

        self.doc_topic_counts_[doc, topic] -= 1
        self.topic_word_counts_[topic, word] -= 1

        return self.doc_topic_counts_, self.topic_word_counts_
    
    def _increment(self, word, doc, topic):
        """TODO
        
        Parameters
        ----------
            word : int
                TODO
            doc : int
                TODO
            topic : int
                TODO

        Returns
        -------
            doc_topic_counts_ : numpy.ndarray, shape=(n_docs, n_topics)
                TODO
            topic_word_counts_ : numpy.ndarray, shape=(n_topics, vocab_size)
                TODO
        """

        self.doc_topic_counts_[doc, topic] += 1
        self.topic_word_counts_[topic, word] += 1

        return self.doc_topic_counts_, self.topic_word_counts_
    
    def _step(self, maximize=False):
        """TODO
        
        Parameters
        ----------
            maximize : bool, default=False
                Optional.
                TODO

        Returns
        -------
            likelihood_avg : float
                TODO
        """

        likelihood_sum = 0

        for i in range(self.corpus_.shape[0]):
            loc, (word, old_doc, old_topic) = self.corpus_[i, :3], self.corpus_[i, 3:].astype(np.int32)
            self._decrement(word, old_doc, old_topic)
            new_doc, new_topic, likelihood = self._sample(loc, word, old_doc, old_topic, return_likelihood=True, maximize=maximize)
            self._increment(word, new_doc, new_topic)
            self.corpus_[i, -2:] = new_doc, new_topic
            likelihood_sum += likelihood

        return likelihood_sum
    
    def fit(self, X, n_steps=500, patience=10, window=1, desc='SLDA', verbose=1, info_rate=10):
        """TODO
        
        Parameters
        ----------
            X : numpy.ndarray, shape=(n_samples, n_features)
                TODO
            n_steps : int, default=1000
                Optional.
                TODO
            patience : int, default=10
                Optional.
                TODO
            window : int, default=1
                Optional.
                TODO
            desc : str, default='SLDA'
                Optional.
                TODO
            verbose : int, default=1
                Optional.
                TODO
            info_rate : int, default=10
                Optional.
                TODO

        Returns
        -------
            self
                I return therefore I am.
        """

        self._build(X)

        for i in tqdm(range(n_steps), desc=desc) if verbose == 1 else range(n_steps):
            likelihood = self._step()
            self.likelihood_log_.append(likelihood)
            self.silhouette_log_.append(silhouette_score(X, self.transform(X)))
            self.patience_log_.append(self._patience(i, window))

            if verbose == 2 and i%info_rate == 0:
                print(self._disp(i, desc))
           
            if sum(self.patience_log_) >= patience:
                break

        return self
    
    def transform(self, _=None):
        """TODO
        
        Parameters
        ----------
            None

        Returns
        -------
            topics : numpy.ndarray, shape=(n_samples,)
                TODO
        """
        
        topics = np.zeros(self.corpus_.shape[0], dtype=np.int32)

        for i in range(self.corpus_.shape[0]):
            word, doc = self.corpus_[i, 3:5].astype(np.int32)
            topics[i] = self._sample_topic(word, doc)

        return topics

class sceLDA(BaseEstimator, TransformerMixin):
    """TODO
    
    Parameters
    ----------
        n_topics : int
            TODO
        hidden_dim : tuple, shape=(n_hidden,), default=(100, 50)
            Optional.
            TODO
        latent_dim : int, default=25
            Optional.
            TODO
        scale : float, default=0.1
            Optional.
            TODO
        batch_size : int, default=None
            Optional.
            TODO
        n_docs : int, default=150
            Optional.
            TODO
        vocab_size : int, default=25
            Optional.
            TODO
        vocab_steps : int, default=10
            Optional.
            TODO
        sigma : float, default=0.1
            Optional.
            TODO
        alpha : float, default=None
            Optional.
            TODO
        beta : float, default=None
            Optional.
            TODO

    Attributes
    ----------
        vae : Class[torch.nn.Module]
            TODO
        slda : Class[sklearn.base.BaseEstimator, sklearn.base.TransformerMixin]
            TODO
    """

    def __init__(self, n_topics, hidden_dim=(100, 50), latent_dim=25, scale=.1, batch_size=None, n_docs=150, vocab_size=25, vocab_steps=10, sigma=.1, alpha=None, beta=None):
        super().__init__()
        
        self.n_topics = n_topics
        self.hidden_dim = hidden_dim
        self.latent_dim = latent_dim
        self.scale = scale
        self.batch_size = batch_size
        self.n_docs = n_docs
        self.vocab_size = vocab_size
        self.vocab_steps = vocab_steps
        self.sigma = sigma
        self.alpha = alpha
        self.beta = beta

        self.vae = None
        self.slda = None

    def _build(self, input_dim):
        """TODO
        
        Parameters
        ----------
            input_dim : int
                TODO
        
        Returns
        -------
            vae : Class[torch.nn.Module]
                TODO
            slda : Class[sklearn.base.BaseEstimator, sklearn.base.TransformerMixin]
                TODO
        """

        if self.vae is None:
            layers = (input_dim,) + self.hidden_dim + (self.latent_dim,)
            self.vae = VAE(*layers, scale=self.scale, batch_size=self.batch_size)

        if self.slda is None:
            self.slda = SLDA(self.n_topics, self.n_docs, self.vocab_size, self.vocab_steps, self.sigma, self.alpha, self.beta)

        return self.vae, self.slda
    
    def tune(self, param_grid, threshold=.99):
        """TODO
        
        Parameters
        ----------
            param_grid : dict
                TODO
            threshold : float, default=0.99
                Optional.
                TODO

        Returns
        -------
            self
                I return therefore I am.
        """


    
    def fit(self, X, n_vae_steps=1000, learning_rate=1e-2, vae_patience=10, vae_window=5, n_slda_steps=500, slda_patience=20, slda_window=5, verbose=1, info_rate=10):
        """TODO
        
        Parameters
        ----------
            X : torch.tensor, shape=(n_samples, n_features)
                TODO
            n_vae_steps : int, default=1000
                Optional.
                TODO
            learning_rate : float, default=1e-2
                Optional.
                TODO
            vae_patience : int, default=10
                Optional.
                TODO
            vae_window : int, default=5
                Optional.
                TODO
            n_slda_steps : int, default=500
                Optional.
                TODO
            slda_patience : int, default=10
                Optional.
                TODO
            slda_window : int, default=5
                Optional.
                TODO
            verbose : int, default=1
                Optional.
                TODO
            info_rate : int, default=10
                Optional.
                TODO

        Returns
        -------
            self
                I return therefore I am.
        """

        if not torch.is_tensor(X):
            X = torch.tensor(X, dtype=torch.float32)

        self._build(X.shape[1] - 3)
        locs, features = X[:, :3], X[:, 3:]
        self.vae.fit(features, n_vae_steps, learning_rate, vae_patience, vae_window, verbose=verbose, info_rate=info_rate)
        Z = torch.hstack([locs, self.vae.transform(features)]).detach().numpy()
        self.slda.fit(Z, n_slda_steps, slda_patience, slda_window, verbose=verbose, info_rate=info_rate)

        return self
    
    def transform(self, X):
        """TODO
        
        Parameters
        ----------
            X : torch.tensor, shape=(n_samples, n_features)
                TODO

        Returns
        -------
            topics : torch.tensor, shape=(n_samples,)
                TODO
        """

        topics = self.slda.transform(X)

        return topics



import numpy as np
import torch
from sklearn.datasets import make_classification
from .util import itemize

HSBLOCKS = np.array([[0, 0],
                     [1, 1]], dtype=np.int32)

VSBLOCKS = np.array([[0, 1],
                     [0, 1]], dtype=np.int32)

CHBLOCKS = np.array([[0, 1, 0],
                     [1, 0, 1],
                     [0, 1, 0]], dtype=np.int32)

GGBLOCKS = np.array([[0, 1, 0, 2, 2, 2],
                     [1, 1, 1, 2, 0, 2],
                     [0, 1, 0, 2, 2, 2],
                     [3, 0, 0, 4, 4, 4],
                     [3, 3, 0, 0, 4, 0],
                     [3, 3, 3, 0, 4, 0]], dtype=np.int32)

def partition(n_samples, n_partitions):
    """TODO
    
    Parameters
    ----------
        n_samples : int
            TODO
        n_partitions : int
            TODO

    Returns
    -------
        parts : list, shape=(n_partitions,)
            TODO
    """

    parts = [n_samples // n_partitions]

    for i in range(1, n_partitions):
        parts.append((n_samples - sum(parts))//(n_partitions - i))
        
    return parts

def parameterize(n_components, n_features, n_equivocal=0, n_redundant=0, mean_scale=1., variance_scale=1.):
    """TODO
    
    Parameters
    ----------
        n_components : int
            TODO
        n_features : int
            TODO
        n_equivocal : int, default=0
            Optional.
            TODO
        n_redundant : int, default=0
            Optional.
            TODO
        mean_scale : float, default=1.0
            Optional.
            TODO
        variance_scale : float, default=1.0
            Optional.
            TODO

    Returns
    -------
        means : numpy.ndarray, shape=(n_components, n_features)
            TODO
        variances : numpy.ndarray, shape=(n_features,)
            TODO
    """
    
    n_informative = n_features - n_equivocal - n_redundant
    means, _ = make_classification(n_samples=n_components, n_features=n_features, n_informative=n_informative, n_redundant=n_redundant, n_classes=n_components, n_clusters_per_class=1, scale=mean_scale)
    variances = variance_scale*np.random.rand(n_features)

    return means, variances

def designate(n_samples, n_components, n_features, n_equivocal=0, n_redundant=0, sample_counts=None, means=None, variances=None, mean_scale=1., variance_scale=1.):
    """TODO
    
    Parameters
    ----------
        n_samples : int
            TODO
        n_components : int
            TODO
        n_features : int
            TODO
        n_equivocal : int, default=0
            Optional.
            TODO
        n_redundant : int, default=0
            Optional.
            TODO
        sample_counts : list, shape=(n_components,), default=None
            Optional.
            TODO
        means : numpy.ndarray, shape=(n_components, n_features), default=None
            Optional.
            TODO
        variances : numpy.ndarray, shape=(n_features,), default=None
            Optional.
            TODO
        mean_scale : float, default=1.0
            Optional.
            TODO
        variance_scale : float, default=1.0
            Optional.
            TODO

    Returns
    -------
        sample_counts : list, shape=(n_components,)
            TODO
        means : numpy.ndarray, shape=(n_components, n_features)
            TODO
        variances : numpy.ndarray, shape=(n_features,)
            TODO
    """
    
    if sample_counts is None:
        sample_counts = partition(n_samples, n_components)
    if means is None:
        means, _ = parameterize(n_components, n_features, n_equivocal, n_redundant, mean_scale, variance_scale)
    if variances is None:
        _, variances = parameterize(n_components, n_features, n_equivocal, n_redundant, mean_scale, variance_scale)

    return sample_counts, means, variances

def sample(n_samples, n_components, n_features, means, variances, component, n_mix=0):
    """TODO
    
    Parameters
    ----------
        n_samples : int
            TODO
        n_components : int
            TODO
        n_features : int
            TODO
        means : numpy.ndarray, shape=(n_components, n_features)
            TODO
        variances : numpy.ndarray, shape=(n_features,)
            TODO
        component : int
            TODO
        n_mix : int, default=0
            Optional.
            TODO

    Returns
    -------
        data : numpy.ndarray, shape=(n_samples, n_features)
            TODO
        labels : numpy.ndarray, shape=(n_samples,)
            TODO
    """
    
    labels = component*np.ones(n_samples, dtype=np.int32)
    data = np.random.normal(means[component], variances, (n_samples, n_features))
    
    mix_idx = np.random.randint(0, n_samples, n_mix)
    labels[mix_idx] = np.random.randint(0, n_components, n_mix)
    data[mix_idx] = np.random.normal(means[labels[mix_idx]], variances, (n_mix, n_features))
    
    return data, labels

def transfigure(blocks):
    """TODO
    
    Parameters
    ----------
        blocks : str
            TODO

    Returns
    -------
        blocks : numpy.ndarray, shape=(y_dim, x_dim)
            TODO
    """

    if isinstance(blocks, str):
        if blocks == 'hsblocks':
            return HSBLOCKS
        if blocks == 'vsblocks':
            return VSBLOCKS
        if blocks == 'chblocks':
            return CHBLOCKS
        if blocks == 'ggblocks':
            return GGBLOCKS
    
        raise NotImplementedError(f'Block palette "{blocks}" not supported.')
    
    return blocks

def interpret(blocks, share_components=False):  
    """TODO
    
    Parameters
    ----------
        blocks : list, shape=(n_blocks,)
            TODO
        share_components : bool, default=False
            TODO

    Returns
    -------
        blocks : list, shape=(n_blocks,)
            TODO
        n_samples : int
            TODO
        n_components : int
            TODO
    """

    n_samples, n_components = [], []

    for idx, b in enumerate(blocks):
        blocks[idx] = b = transfigure(b)
        n_samples.append(b.shape[0])
        n_components.append(np.unique(b).shape[0])

    n_samples = sum(n_samples)
    n_components = max(n_components) if share_components else sum(n_components)

    return blocks, n_samples, n_components

def make_data(n_samples, n_components, n_features, n_equivocal=0, n_redundant=0, sample_counts=None, means=None, variances=None, mean_scale=1., variance_scale=1., mix=0.):
    """TODO
    
    Parameters
    ----------
        n_samples : int
            TODO
        n_components : int
            TODO
        n_features : int
            TODO
        n_equivocal : int, default=0
            Optional.
            TODO
        n_redundant : int, default=0
            Optional.
            TODO
        sample_counts : list, default=None
            Optional.
            TODO
        means : numpy.ndarray, shape=(n_components, n_features), default=None
            Optional.
            TODO
        variances : numpy.ndarray, shape=(n_features,), default=None
            Optional.
            TODO
        mean_scale : float, default=1.0
            Optional.
            TODO
        variance_scale : float, default=1.0
            Optional.
            TODO
        mix : float, default=0.0
            Optional.
            TODO

    Returns
    -------
        data : numpy.ndarray, shape=(n_samples, n_features)
            TODO
        labels : numpy.ndarray, shape=(n_samples,)
            TODO
    """
    
    n_mix = partition(int(mix*n_samples), n_components)
    sample_counts, means, variances = designate(n_samples, n_components, n_features, n_equivocal, n_redundant, sample_counts, means, variances, mean_scale, variance_scale)
    data = np.zeros((n_samples, n_features), dtype=np.float32)
    labels = np.zeros(n_samples, dtype=np.int32)
    j = 0

    for i in range(n_components):
        k = j + sample_counts[i]
        data[j:k], labels[j:k] = sample(sample_counts[i], n_components, n_features, means, variances, i, n_mix[i])
        j += sample_counts[i]

    return data, labels

def make_blocks(blocks, n_features, n_equivocal=0, n_redundant=0, means=None, variances=None, mean_scale=1., variance_scale=1., block_size=5, wiggle=0., mix=0.):
    """TODO
    
    Parameters
    ----------
        blocks : numpy.ndarray, shape=(y_dim, x_dim)
            TODO
        n_features : int
            TODO
        n_equivocal : int, default=0
            Optional.
            TODO
        n_redundant : int, default=0
            Optional.
            TODO
        means : numpy.ndarray, shape=(n_components, n_features), default=None
            Optional.
            TODO
        variances : numpy.ndarray, shape=(n_features,), default=None
            Optional.
            TODO
        mean_scale : float, default=1.0
            Optional.
            TODO
        variance_scale : float, default=1.0
            Optional.
            TODO
        block_size : int, default=5
            Optional.
            TODO
        wiggle : float, default=0.0
            Optional.
            TODO
        mix : float, default=0.0
            Optional.
            TODO

    Returns
    -------
        data : numpy.ndarray, shape=(n_samples, n_features)
            TODO
        labels : numpy.ndarray, shape=(n_samples,)
            TODO
    """
    
    grid = np.repeat(np.repeat(blocks, block_size, 0), block_size, 1)
    n_samples = grid.shape[0]*grid.shape[1]
    _, sample_counts = np.unique(grid, return_counts=True)
    n_components = sample_counts.shape[0]
    data, labels = make_data(n_samples, n_components, n_features, n_equivocal, n_redundant, sample_counts, means, variances, mean_scale, variance_scale, mix)
    data = np.hstack([np.zeros((data.shape[0], 2)), data])
    j = 0

    for i in range(n_components):
        locs = np.stack(np.where(grid.T == i)).T
        locs[:, -1] = -locs[:, -1] + grid.shape[0]
        data[j:j + locs.shape[0], :2] = locs
        j += locs.shape[0]

    data[:, :2] += 2*wiggle*np.random.rand(*data[:, :2].shape) - wiggle
    
    return data, labels

def make_dataset(blocks, n_features, n_equivocal=0, n_redundant=0, means=None, variances=None, mean_scale=1., variance_scale=1., block_size=5, wiggle=0., mix=0., img=None, offset=0, tensor=False):
    """TODO
    
    Parameters
    ----------
        blocks : numpy.ndarray, shape=(y_dim, x_dim)
            TODO
        n_features : int
            TODO
        n_equivocal : int, default=0
            Optional.
            TODO
        n_redundant : int, default=0
            Optional.
            TODO
        means : numpy.ndarray, shape=(n_components, n_features), default=None
            Optional.
            TODO
        variances : numpy.ndarray, shape=(n_features,), default=None
            Optional.
            TODO
        mean_scale : float, default=1.0
            Optional.
            TODO
        variance_scale : float, default=1.0
            Optional
            TODO
        block_size : int, default=5
            Optional.
            TODO
        wiggle : float, default=0.0
            Optional.
            TODO
        mix : float, default=0.0
            Optional.
            TODO
        img : int, default=None
            Optional.
            TODO
        offset : int, default=0
            Optional.
            TODO
        tensor : bool, default=False
            Optional.
            TODO

    Returns
    -------
        data : numpy.ndarray|torch.Tensor, shape=(n_samples, n_features)
            TODO
        labels : numpy.ndarray|torch.Tensor, shape=(n_samples,)
            TODO
    """

    if means is not None:
        means = means[offset:]
    
    blocks = transfigure(blocks)
    data, labels = make_blocks(blocks, n_features, n_equivocal, n_redundant, means, variances, mean_scale, variance_scale, block_size, wiggle, mix)
    labels += offset

    if img is not None:
        data = np.hstack([img*np.ones((data.shape[0], 1)), data])

    if tensor:
        return torch.tensor(data, dtype=torch.float32), torch.tensor(labels, dtype=torch.int32)
    return data, labels

def make_datasets(blocks, n_features, n_equivocal=0, n_redundant=0, means=None, variances=None, mean_scale=1., variance_scale=1., block_size=5, wiggle=0., mix=0., share_components=True, tensor=False):
    """TODO
    
    Parameters
    ----------
        blocks : list|numpy.ndarray, shape=(n_datasets,)|(y_dim, x_dim)
            TODO
        n_features : int
            TODO
        n_equivocal : int, default=0
            Optional.
            TODO
        n_redundant : int, default=0
            Optional.
            TODO
        means : numpy.ndarray, shape=(n_components, n_features), default=None
            Optional.
            TODO
        variances : numpy.ndarray, shape=(n_features,), default=None
            Optional.
            TODO
        mean_scale : float, default=1.0
            Optional.
            TODO
        variance_scale : float, default=1.0
            Optional
            TODO
        block_size : int|list, default=5
            Optional.
            TODO
        wiggle : float|list, default=0.0
            Optional.
            TODO
        mix : float|list, default=0.0
            Optional.
            TODO
        share_components : bool, default=True
            Optional.
            TODO
        tensor : bool, default=False
            Optional.
            TODO

    Returns
    -------
        data : numpy.ndarray|torch.Tensor, shape=(n_samples, n_features)
            TODO
        labels : numpy.ndarray|torch.Tensor, shape=(n_samples,)
            TODO
    """
    
    n_datasets = len(blocks) if isinstance(blocks, list) else 1
    blocks, block_size, wiggle, mix = itemize(n_datasets, blocks, block_size, wiggle, mix)
    blocks, _, n_components = interpret(blocks, share_components)
    _, means, variances = designate(_, n_components, n_features, n_equivocal, n_redundant, _, means, variances, mean_scale, variance_scale)
    datasets = []

    for idx, (b, s, w, m) in enumerate(zip(blocks, block_size, wiggle, mix)):
        offset = 0 if share_components or len(datasets) == 0 else max(datasets[-1][1]) + 1
        datasets.append(make_dataset(b, n_features, n_equivocal, n_redundant, means, variances, mean_scale, variance_scale, s, w, m, idx, offset))

    data, labels = zip(*datasets)
    data = np.concatenate(data)
    labels = np.concatenate(labels)

    if tensor:
        return torch.tensor(data, dtype=torch.float32), torch.tensor(labels, dtype=torch.int32)
    return data, labels



import pyro
import pyro.distributions as dist
import torch
import torch.nn as nn
from pyro.infer import SVI, Trace_ELBO
from pyro.optim import Adam
from tqdm import tqdm
from .util import set_seed

def activation(act='relu', **kwargs):
    """Instantiates and returns the specified torch.nn activation object.
    
    Parameters
    ----------
        act : str, default='relu'
            Optional.
            Activation to instantiate.
            Supports 'relu', 'softplus', 'sigmoid', 'softmax' activations.
        kwargs : dict
            Optional.
            Additional arguments needed for the specified activation.

    Raises
    ------
        NotImplementedError
            If the specified activation is not supported.

    Returns
    -------
        activation: torch.nn.*
            Instantiated object corresponding to the specified activation.
    """

    if act == 'relu':
        return nn.ReLU(**kwargs)
    elif act == 'softplus':
        return nn.Softplus(**kwargs)
    elif act == 'sigmoid':
        return nn.Sigmoid(**kwargs)
    elif act == 'softmax':
        return nn.Softmax(dim=-1)
    
    raise NotImplementedError(f'Activation function"{act}" not supported.')

def mlp(layers, bias=True, act='relu', final_act=None, batch_norm=True, affine=True, dropout=0.):
    """Generates linear torch.nn layers with activation, dropout, and optional
    batch normalization.
    
    Parameters
    ----------
        layers : tuple, shape=(n_layers,)
            Sequence of integer values specifying the size of each layer.
        bias : bool, default=True
            Optional.
            Whether or not to learn an additive bias at each layer.
        act : str, default='relu'
            Optional.
            Hidden activation function.
            Supports 'relu', 'softplus', 'sigmoid', and 'softmax' activations.
        final_act : str, default=None
            Optional.
            Final activation function.
            Supports 'softplus', 'relu', 'sigmoid', and 'softmax' activations.
        batch_norm : bool, default=True
            Optional.
            Whether or not to perform final batch normalization.
        affine : bool, default=True
            Optional.
            Whether or not to allow for learnable affine parameters.
        dropout : float, default=0.0
            Optional.
            Probability of random node omission during training.

    Returns
    -------
        mlp : generator
            Iterable object that constructs the specified linear neural network.
    """
    
    n_layers = len(layers) - 1

    for i in range(1, n_layers + 1):
        yield nn.Linear(layers[i - 1], layers[i], bias=bias)
        if i < n_layers - 1:
            yield activation(act)
        else:
            if batch_norm:
                yield nn.BatchNorm1d(layers[i], affine=affine)
            if final_act:
                yield activation(final_act)
            yield nn.Dropout(dropout)

class MLP(nn.Module):
    """TODO

    Parameters
    ----------
        layers : tuple, shape=(n_layers,)
            Sequence of integer values specifying the size of each layer.
        bias : bool, default=True
            Optional.
            Whether or not to learn an additive bias at each layer.
        act : str, default='relu'
            Optional.
            Hidden activation function.
            Supports 'softplus', 'relu', 'sigmoid', and 'softmax' activations.
        final_act : str, default=None
            Optional.
            Final activation function.
            Supports 'softplus', 'relu', 'sigmoid', and 'softmax' activations.
        batch_norm : bool, default=True
            Optional.
            Whether or not to perform final batch normalization.
        affine : bool, default=True
            Optional.
            Whether or not to allow for learnable affine parameters.
        dropout : float, default=0.0
            Optional.
            Probability of random node omission during training.
        
    Attributes
    ----------
        net : torch.nn.Sequential
            TODO
    """

    def __init__(self, layers, bias=True, act='relu', final_act=None, batch_norm=True, affine=True, dropout=0.):
        super().__init__()

        self.net = nn.Sequential(*list(mlp(layers, bias, act, final_act, batch_norm, affine, dropout)))
    
    def forward(self, x):
        """TODO
        
        Parameters
        ----------
            x : torch.Tensor, shape=(batch_size, input_dim)
                TODO

        Returns
        -------
            y : torch.Tensor, shape=(batch_size, output_dim)
                TODO
        """

        y = self.net(x)

        return y

class Encoder(nn.Module):
    """TODO

    Parameters
    ----------
        layers : tuple, shape=(n_layers)
            TODO
        
    Attributes
    ----------
        encoding_net : Class[torch.nn.Module]
            TODO
        loc_net : Class[torch.nn.Module]
            TODO
        scale_net : Class[torch.nn.Module]
            TODO
    """
    
    def __init__(self, *layers):
        super().__init__()

        encoding_layers = layers[:-1]
        self.encoding_net = MLP(encoding_layers, final_act='softplus', batch_norm=False, dropout=.2)

        latent_layers = layers[-2:]
        self.loc_net, self.scale_net = MLP(latent_layers), MLP(latent_layers)

    def forward(self, x):
        """TODO
        
        Parameters
        ----------
            x : torch.Tensor, shape=(batch_size, input_dim)
                TODO
        
        Returns
        -------
            z_loc : torch.Tensor, shape=(batch_size, latent_dim)
                TODO
            z_scale : torch.Tensor, shape=(batch_size, latent_dim)
                TODO
        """
        
        y = self.encoding_net(x)
        z_loc = self.loc_net(y)
        z_scale = (.5*self.scale_net(y)).exp()

        return z_loc, z_scale
    
class Decoder(nn.Module):
    """TODO
    
    Parameters
    ----------
        layers : tuple, shape=(n_layers)
            TODO
        
    Attributes
    ----------
        net : Class[torch.nn.Module]
            TODO
    """

    def __init__(self, *layers):
        super().__init__()

        self.net = MLP(layers)

    def forward(self, z):
        """TODO
        
        Parameters
        ----------
            z : torch.Tensor, shape=(batch_size, latent_dim)
                TODO

        Returns
        -------
            x : torch.Tensor, shape=(batch_size, output_dim)
                TODO
        """

        x = self.net(z)

        return x

class VAE(nn.Module):
    """TODO
    
    Parameters
    ----------
        layers : tuple, shape=(n_layers)
            TODO
        scale : float, default=1.0
            Optional.
            TODO
        batch_size : int, default=None
            Optional.
            TODO
        seed : int, default=9
            Optional.
            TODO
        
    Attributes
    ----------
        encoder : Class[torch.nn.Module]
            TODO
        decoder : Class[torch.nn.Module]
            TODO
        training_log_ : list, shape=(n_steps,)
            TODO
        validation_log_ : list, shape=(n_steps,)
            TODO
    """

    def __init__(self, *layers, scale=.1, batch_size=None, seed=9):
        super().__init__()
        pyro.clear_param_store()

        if seed is not None:
            set_seed(seed)

        self.layers = layers
        self.scale = scale
        self.batch_size = batch_size

        self.encoder = Encoder(*layers)
        self.decoder = Decoder(*layers[::-1])
        self.training_log_ = []
        self.validation_log_ = []
        self.patience_log_ = []

    def _disp(self, idx, desc='VAE'):
        """TODO
        
        Parameters
        ----------
            idx : int
                TODO
            desc : str, default='VAE'

        Returns
        -------
            disp : str
                TODO
        """

        scores = self.training_log_[idx], self.validation_log_[idx], sum(self.patience_log_[idx])
        disp = f'{desc} | step {idx} | training loss: {scores[0]} | validation loss: {scores[1]} | patience: {scores[2]}'

        return disp
    
    def _build(self, learning_rate):
        """TODO
        
        Parameters
        ----------
            learning_rate : float
                TODO

        Returns
        -------
            elbo : TODO
                TODO
            svi : TODO
                TODO
        """

        optim = Adam({'lr': learning_rate})
        elbo = Trace_ELBO(max_iarange_nesting=4)
        svi = SVI(self._model, self._guide, optim, elbo)

        return elbo, svi

    def _patience(self, idx, window):
        """TODO
        
        Parameters
        ----------
            idx : int
                TODO
            window : int
                TODO

        Returns
        -------
            patience : bool
                TODO
        """

        if idx%window == 0 and idx > 2*window:
            return sum(self.validation_log_[-window:]) >= sum(self.validation_log_[-2*window:-window])
        return False
    
    def _model(self, x):
        """TODO
        
        Parameters
        ----------
            x : torch.Tensor, shape=(n_samples, input_dim)
                TODO

        Returns
        -------
            None
        """
        
        pyro.module('decoder', self.decoder)

        with pyro.plate('data', x.shape[0], self.batch_size) as idx:
            z_loc = x[idx].new_zeros((x[idx].shape[0], self.layers[-1]))
            z_scale = x[idx].new_ones((x[idx].shape[0], self.layers[-1]))
            z = pyro.sample('latent', dist.Normal(z_loc, z_scale).to_event(1))
            x_loc = self.decoder(z)
            x_scale = self.scale*torch.ones_like(x_loc)
            pyro.sample('obs', dist.Normal(x_loc, x_scale).to_event(1), obs=x[idx])

    def _guide(self, x):
        """TODO
        
        Parameters
        ----------
            x : torch.Tensor, shape=(n_samples, input_dim)
                TODO

        Returns
        -------
            None
        """

        pyro.module('encoder', self.encoder)

        with pyro.plate('data', x.shape[0], self.batch_size) as idx:
            z_loc, z_scale = self.encoder(x[idx])
            pyro.sample('latent', dist.Normal(z_loc, z_scale).to_event(1))

    def fit(self, X, n_steps=1000, learning_rate=1e-2, patience=10, window=5, val_split=.2, desc='VAE', verbose=1, info_rate=10):
        """TODO
        
        Parameters
        ----------
            X : torch.Tensor, shape=(n_samples, input_dim)
                Dataset with which to train the networks.
            n_steps : int, default=1000
                Optional.
                Number of steps to train the networks for.
            learning_rate : float, default=1e-2
                Optional.
                Step size for gradient descent during training.
            patience : int, default=10
                Optional.
                TODO
            window : int, default=5
                Optional.
                TODO
            val_split : float, default=0.2
                Optional.
                TODO
            desc : str, default='VAE'
                Optional.
                Description displayed with information reported during training.
            verbose : int, default=1
                Optional.
                Level of verbosity of information reported during training.
                Supports 0, 1, and 2.
            info_rate : int, default=10
                Optional.
                TODO

        Returns
        -------
            self
                I return therefore I am.
        """

        idx = torch.randperm(X.shape[0])
        split_idx = int(val_split*X.shape[0])
        X_train, X_val = X[idx[split_idx:]], X[idx[:split_idx]]
        elbo, svi = self._build(learning_rate)
        
        for i in tqdm(range(n_steps), desc=desc) if verbose == 1 else range(n_steps):
            training_loss = svi.step(X_train)

            with torch.no_grad():
                self.training_log_.append(training_loss)
                self.validation_log_.append(elbo.loss(self._model, self._guide, X_val))
                self.patience_log_.append(self._patience(i, window))

                if verbose == 2 and i%info_rate == 0:
                    print(self._disp(i, desc))

                if sum(self.patience_log_) >= patience:
                    break

        return self
    
    def transform(self, X):
        """TODO
        
        Parameters
        ----------
            X : torch.Tensor, shape=(n_samples, input_dim)
                Dataset with which the networks were trained.
            
        Returns
        -------
            Z : torch.Tensor, shape=(n_samples, latent_dim)
                Dataset sampled from the latent space encoded with X
        """

        Z_loc, Z_scale = self.encoder(X)
        Z = dist.Normal(Z_loc, Z_scale).sample()

        return Z
